﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Irony.Parsing;

namespace searchComponent.grammar
{
    /// <summary>
    /// This class represents the date Grammar, this is used only for dates.
    /// Some examples are: 21/05/1987, 26/03/1994:27/03/1994, &lt;26/04/2010,
    /// </summary>
    public class DateGrammar : Grammar
    {
        /// <summary>
        /// This is where the grammar is defined.
        /// </summary>
        public DateGrammar()
        {
            this.GrammarComments = @"
Simple grammar to operate on dates. It allows dates and date comparisons like <date
or !date, also it allows and or operators and the : operator which means between. ";

            // 1. Terminals
            var day = new NumberLiteral("day");
            day.Options = NumberOptions.IntOnly;
            day.ValidateToken += new EventHandler<ParsingEventArgs>(day_ValidateToken);
            var month = new NumberLiteral("month");
            month.Options = NumberOptions.IntOnly;
            month.ValidateToken += new EventHandler<ParsingEventArgs>(month_ValidateToken);
            var year = new NumberLiteral("year");
            year.Options = NumberOptions.IntOnly;
            year.ValidateToken += new EventHandler<ParsingEventArgs>(year_ValidateToken);
            var ImpliedAnd = new ImpliedSymbolTerminal("ImpliedAnd"); 

            // 2. Non Terminals
            var SingleTerm = new NonTerminal("SingleTerm");
            var DateTerm = new NonTerminal("DateTerm");
            var UnaryTerm = new NonTerminal("UnaryTerm");
            var UnaryOp = new NonTerminal("UnaryOp");
            var BinaryTerm = new NonTerminal("BinaryTerm");
            var BinaryOp = new NonTerminal("BinaryOp");
            var Operand = new NonTerminal("Operand");
            var AndExpression = new NonTerminal("AndExpression");
            var AndTerm = new NonTerminal("AndTerm");
            var OrExpression = new NonTerminal("OrExpression");
            var OrOp = new NonTerminal("OrOp");
            var AndOp = new NonTerminal("AndOp");
            var Expression = new NonTerminal("Expression");

            // 3. BNF Rules
            Expression.Rule = OrExpression | AndExpression | Operand;
            OrExpression.Rule = Expression + OrOp + AndTerm;
            AndExpression.Rule = AndTerm + AndOp + Operand;
            AndTerm.Rule = AndExpression | Operand;

            OrOp.Rule = ToTerm("or");
            AndOp.Rule = ToTerm("and") | ImpliedAnd;

            Operand.Rule = UnaryTerm | BinaryTerm;
            UnaryTerm.Rule = (UnaryOp + SingleTerm) | (UnaryOp + DateTerm) | DateTerm | SingleTerm;
            SingleTerm.Rule = 
                (ToTerm("d") + "(" + day + ")") |
                (ToTerm("m") + "(" + month + ")") |
                (ToTerm("y") + "(" + year + ")");
            BinaryTerm.Rule = DateTerm + BinaryOp + DateTerm;
            UnaryOp.Rule = ToTerm("=") | ">" | "<" | "!";
            BinaryOp.Rule = ToTerm(":");

            DateTerm.Rule = day + "/" + month + "/" + year;

            this.Root = Expression;

            // 4. Operators precedence
            RegisterOperators(10, AndOp);
            RegisterOperators(20, OrOp);
        }

        /// <summary>
        /// This validator will allow the insertion of numbers between 1000 and 3000
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        void year_ValidateToken(object sender, ParsingEventArgs e)
        {
            int yy = Convert.ToInt32(e.Context.CurrentToken.Text);
            if (yy < 1790 || yy > 9999)
            {
                e.Context.CurrentToken = e.Context.Source.CreateErrorToken("valid years between 1790 and 9999");
            }
        }

        /// <summary>
        /// This validator will allow the insertion of numbers between 1 and 31
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        void day_ValidateToken(object sender, ParsingEventArgs e)
        {
            int dd = Convert.ToInt32(e.Context.CurrentToken.Text);
            if (dd < 1 || dd > 31)
            {
                e.Context.CurrentToken = e.Context.Source.CreateErrorToken("invalid day");
            }
        }

        /// <summary>
        /// This validator will allow the insertion of numbers between 1 and 12
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        void month_ValidateToken(object sender, ParsingEventArgs e)
        {
            int mm = Convert.ToInt32(e.Context.CurrentToken.Text);
            if (mm < 1 || mm > 12)
            {
                e.Context.CurrentToken = e.Context.Source.CreateErrorToken("invalid month");
            }
        }
    }
}
