﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using searchComponent.structure;
using System.Globalization;
using Irony.Parsing;
using Irony.Interpreter;

namespace searchComponent.grammar
{
    /// <summary>
    /// This class represents the field grammar, this is dinamycally constructed
    /// from the SearchStructure defined.
    /// </summary>
    public class FieldGrammar : Grammar
    {
        #region Fields
        /// <summary>
        /// This is the structure containing the rules, tables and others that define the
        /// search to be executed.
        /// </summary>
        private SearchStructure _structure;
        #endregion 

        #region Properties
        /// <summary>
        /// this property allows set and get methods because the client could change
        /// the purpose of the searchcompoennte any time.
        /// </summary>
        public SearchStructure Structure
        {
            get { return this._structure; }
            set { this._structure = value; }
        }
        #endregion

        #region Constructor
        /// <summary>
        /// This is where the grammar is defined. By default it will load the confifguration
        /// for test1
        /// </summary>
        public FieldGrammar() : this( SearchStructure.GetStructure("test1", SearchStructure.DEFAULT_CONN_STRING))
        {
            ;
        }

        /// <summary>
        /// This is the constructor that takes a searchstructure as a parameter and loads
        /// all needed to form the grammar.
        /// </summary>
        /// <param name="objStructure"></param>
        public FieldGrammar(SearchStructure objStructure)
        {
            _structure = objStructure;

            this.GrammarComments =
            "Simple grammar: only terms and phrases are recognized";

            // 1. Terminals
            var Name = new IdentifierTerminal("Name");
            //Name.Priority = Terminal.LowestPriority;
            Name.ValidateToken += new EventHandler<ParsingEventArgs>(Name_ValidateToken);

            var Term = CreateTerm("Term");
            Term.Priority = Terminal.LowestPriority;

            // 2. Non Terminals
            var Expression = new NonTerminal("Expression");
            var ColumnSearch = new NonTerminal("ColumnSearch");
            var QuerySearchOp = new NonTerminal("QuerySearchOp");
            var EqualOp = new NonTerminal("EqualOp");
            var AndOp = new NonTerminal("AndOp");
            var OrOp = new NonTerminal("OrOp");

            // 3. BNF
            Expression.Rule = ColumnSearch | (Expression + QuerySearchOp + ColumnSearch);
            ColumnSearch.Rule = Name + EqualOp + Term;
            QuerySearchOp.Rule = AndOp | OrOp;
            AndOp.Rule = ToTerm("&");
            OrOp.Rule = ToTerm("|");            
            EqualOp.Rule = "=";

            this.Root = Expression;

            // 4. Operators precedence
            RegisterOperators(30, EqualOp);
            RegisterOperators(20, OrOp);
            RegisterOperators(10, AndOp);
        }
        #endregion

        #region Methods
        /// <summary>
        /// Validates that the name for the token is only the tags allowed
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        void Name_ValidateToken(object sender, ParsingEventArgs e)
        {
            bool foundName = false;
            foreach (TableStructure tbl in this._structure.Tables)
            {
                foreach (Column col in tbl.Columns)
                {
                    if (col.Searchable)
                    {
                        foundName = col.SearchKey.Equals(e.Context.CurrentToken.Text);
                    }
                    if (foundName) break;
                }
                if (foundName) break;
            }

            if (!foundName && !e.Context.CurrentToken.Text.Equals("all"))
            {
                e.Context.CurrentToken = 
                    e.Context.Source.CreateErrorToken("Not a valid token: " + e.Context.CurrentToken.Text);
            }
        }

        //Creates extended identifier terminal that allows international characters
        // Following the pattern used for c# identifier terminal in TerminalFactory.CreateCSharpIdentifier method;
        private IdentifierTerminal CreateTerm(string name)
        {
            IdentifierTerminal term = new IdentifierTerminal(name, "!@#$%^*_\"'.?-:{}()<>/ ", "!@#$%^*_\"'.?0123456789{}()<>");
            term.CharCategories.AddRange(new UnicodeCategory[] {
             UnicodeCategory.UppercaseLetter, //Ul
             UnicodeCategory.LowercaseLetter, //Ll
             UnicodeCategory.TitlecaseLetter, //Lt
             UnicodeCategory.ModifierLetter,  //Lm
             UnicodeCategory.OtherLetter,     //Lo
             UnicodeCategory.LetterNumber,     //Nl
             UnicodeCategory.DecimalDigitNumber, //Nd
             UnicodeCategory.ConnectorPunctuation, //Pc
             UnicodeCategory.SpacingCombiningMark, //Mc
             UnicodeCategory.NonSpacingMark,       //Mn
             UnicodeCategory.Format                //Cf
          });
            //StartCharCategories are the same
            term.StartCharCategories.AddRange(term.CharCategories);
            return term;
        }
        #endregion
    }
}
