﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using searchComponent.structure;
using Irony.Parsing;
using searchComponent.grammar;

namespace searchComponent.interpreter
{
    /// <summary>
    /// This class will form the sql query for the date columns
    /// </summary>
    public class DateColumnInterpreter : ColumnInterpreter
    {
        #region Constructor 
        /// <summary>
        /// Constructs a DateColumnInterpreter linked to a specific column and ready to 
        /// interpret a specific query.
        /// </summary>
        /// <param name="objColumn"></param>
        public DateColumnInterpreter(Column objColumn)
        {
            this._column = objColumn;
        }
        #endregion

        #region Methods
        /// <summary>
        /// Gets the query needed for this column. THe node is present as a parameter
        /// because the method is recursive.
        /// </summary>
        /// <param name="node"></param>
        /// <returns></returns>
        public override string GetWhereQuery(string columnSearch)
        {
            DateGrammar grammar = new DateGrammar();
            LanguageData language = new LanguageData(grammar);
            Parser parser = new Parser(language);
            ParseTree parseTree = parser.Parse(columnSearch);
            if (parseTree.Status == ParseTreeStatus.Error)
            {
                string messages = "";
                ParserMessageList listMessages = parseTree.ParserMessages;
                foreach (ParserMessage objMsg in listMessages)
                {
                    messages += objMsg.Message + "\n";
                }
                throw new Exception(messages);
            }
            ParseTreeNode root = parseTree.Root;
            string query = ConvertQuery(root);
            return query;
        }

        /// <summary>
        /// This method should convert the node tree into a sql query for the column passed
        /// as a parameter in the constructor of this object.
        /// </summary>
        /// <param name="node"></param>
        /// <returns></returns>
        private string ConvertQuery(ParseTreeNode node)
        {
            string result = "";
            // Note that some NonTerminals don't actually get into the AST tree, 
            // because of some Irony's optimizations - punctuation stripping and 
            // node bubbling. For example, ParenthesizedExpression - parentheses 
            // symbols get stripped off as punctuation, and child expression node 
            // (parenthesized content) replaces the parent ParExpr node (the 
            // child is "bubbled up").
            switch (node.Term.Name)
            {
                case "Expression":
                    result += ConvertQuery(node.ChildNodes[0]);
                    break;

                case "OrExpression":
                    result += "( " + ConvertQuery(node.ChildNodes[0]) + " )";
                    result += " OR (" + ConvertQuery(node.ChildNodes[2]) + " )";
                    break;

                case "AndExpression":
                    result += "( " + ConvertQuery(node.ChildNodes[0]) + " )";
                    result += " AND (" + ConvertQuery(node.ChildNodes[2]) + " )";
                    break;

                case "AndTerm":
                    result += ConvertQuery(node.ChildNodes[0]);
                    break;

                case "Operand":
                    result += ConvertQuery(node.ChildNodes[0]);
                    break;

                case "UnaryTerm":
                    // Can be a single date or a singleterm (only day or month)
                    if (node.ChildNodes.Count == 1)
                    {
                        result += ConvertColumnName(node.ChildNodes[0]) + "=" + ConvertQuery(node.ChildNodes[0]);
                        break;
                    }
                    // operation + date or singleterm
                    result += ConvertColumnName(node.ChildNodes[1]) + " " +
                        GetComparisonOperator(node.ChildNodes[0].ChildNodes[0].Token.Text) + " " +
                        ConvertQuery(node.ChildNodes[1]);
                    break;
                case "BinaryTerm":
                    result += this._column.FullName + " between " + ConvertQuery(node.ChildNodes[0]) +
                        " and " + ConvertQuery(node.ChildNodes[2]);
                    break;
                case "DateTerm":
                    result += ToSafeSqlDateValue(
                        node.ChildNodes[0].Token.Text,
                        node.ChildNodes[2].Token.Text,
                        node.ChildNodes[4].Token.Text);
                    break;
                case "SingleTerm":
                    result += node.ChildNodes[2].Token.Text;
                    break;
                // This should never happen, even if input string is garbage
                default:
                    result += " Please investigate.";
                    break;
            }
            return result;
        }

        /// <summary>
        /// This method converts the tree node in the corresponding syntax for comparison with
        /// the date. If comparison in the where clause is done with DateTerm then we return 
        /// the fullname of the column, otherwise, return the dya, month or year operation over the
        /// fullname of the column
        /// </summary>
        /// <param name="parseTreeNode"></param>
        /// <returns></returns>
        private string ConvertColumnName(ParseTreeNode node)
        {
            switch (node.Term.Name)
            {
                case "DateTerm":
                    return this._column.FullName;
                case "SingleTerm":
                    return ConvertColumnName(node.ChildNodes[2]);
                case "day":
                    return "DAY(" + this._column.FullName + ")";
                case "month":
                    return "MONTH(" + this._column.FullName + ")";
                case "year":
                    return "YEAR(" + this._column.FullName + ")";
                default:
                    return "Please investigate.";
            }
        }

        /// <summary>
        /// Converts the parameters into a date safe value ready to be used in a sql Query.
        /// We are using the formula in: http://weblogs.sqlteam.com/jeffs/archive/2007/01/02/56079.aspx
        /// where date is used as:
        /// dateadd(month,((@Year-1900)*12)+@Month-1,@Day-1)
        /// </summary>
        /// <param name="day"></param>
        /// <param name="month"></param>
        /// <param name="year"></param>
        /// <returns></returns>
        private string ToSafeSqlDateValue(string day, string month, string year)
        {
            return "dateadd(month,((" + year + "-1900)*12)+" + month + "-1," + day + "-1)";
        }

        /// <summary>
        /// Retiurns the comparison operator to be used. Actually, all is needed is to
        /// return the same operator than the grammar and in the case of ! we have 
        /// to return &lt;&gt;
        /// </summary>
        /// <param name="p"></param>
        /// <returns></returns>
        private string GetComparisonOperator(string p)
        {
            if (p.Equals("!"))
                return "<>";
            return p;
        }
        #endregion
    }
}
