﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using searchComponent.structure;
using Irony.Parsing;
using searchComponent.grammar;

namespace searchComponent.interpreter
{
    /// <summary>
    /// This class will form the sql query for the Fts columns
    /// </summary>
    public class FtsColumnInterpreter : ColumnInterpreter
    {
        #region Enums
        /// <summary>
        /// The type of searches that can be performed with FTS
        /// </summary>
        public enum TermType
        {
            Inflectional = 1,
            Thesaurus = 2,
            Exact = 3
        };
        #endregion

        #region Constructor
        /// <summary>
        /// Constructs a NumericColumnInterpreter linked to a specific column and ready to 
        /// interpret a specific query.
        /// </summary>
        /// <param name="objColumn"></param>
        public FtsColumnInterpreter(Column objColumn)
        {
            this._column = objColumn;
        }
        #endregion

        #region Methods
        /// <summary>
        /// Gets the query needed for this column. THe node is present as a parameter
        /// because the method is recursive.
        /// </summary>
        /// <param name="node"></param>
        /// <returns></returns>
        public override string GetWhereQuery(string columnSearch)
        {
            FtsGrammar grammar = new FtsGrammar();
            LanguageData language = new LanguageData(grammar);
            Parser parser = new Parser(language);
            ParseTree parseTree = parser.Parse(columnSearch);
            if (parseTree.Status == ParseTreeStatus.Error)
            {
                string messages = "";
                ParserMessageList listMessages = parseTree.ParserMessages;
                foreach (ParserMessage objMsg in listMessages)
                {
                    messages += objMsg.Message + "\n";
                }
                throw new Exception(messages);
            }
            ParseTreeNode root = parseTree.Root;
            string query = "CONTAINS(" + _column.FullName + ",'" + ConvertQuery(root) + "')";
            return query;
        }

        /// <summary>
        /// This method should convert the node tree into a sql query for the column passed
        /// as a parameter in the constructor of this object.
        /// </summary>
        /// <param name="node"></param>
        /// <returns></returns>
        private string ConvertQuery(ParseTreeNode node)
        {
            return ConvertQuery(node, TermType.Inflectional); 
        }

        /// <summary>
        /// This method should convert the node tree into a sql query for the column passed
        /// as a parameter in the constructor of this object.
        /// </summary>
        /// <param name="node"></param>
        /// <returns></returns>
        private string ConvertQuery(ParseTreeNode node, TermType type)
        {
            string result = "";
            // Note that some NonTerminals don't actually get into the AST tree, 
            // because of some Irony's optimizations - punctuation stripping and 
            // transient nodes elimination. For example, ParenthesizedExpression - parentheses 
            // symbols get stripped off as punctuation, and child expression node 
            // (parenthesized content) replaces the parent ParenthesizedExpression node
            switch (node.Term.Name)
            {
                case "BinaryExpression":
                    string opSym = string.Empty;
                    string op = node.ChildNodes[1].FindTokenAndGetText().ToLower();
                    string sqlOp = "";
                    switch (op)
                    {
                        case "":
                        case "&":
                        case "and":
                            sqlOp = " AND ";
                            type = TermType.Inflectional;
                            break;
                        case "-":
                            sqlOp = " AND NOT ";
                            break;
                        case "|":
                        case "or":
                            sqlOp = " OR ";
                            break;
                    }//switch

                    result = "(" + ConvertQuery(node.ChildNodes[0], type) + sqlOp + ConvertQuery(node.ChildNodes[2], type) + ")";
                    break;

                case "PrimaryExpression":
                    result = "(" + ConvertQuery(node.ChildNodes[0], type) + ")";
                    break;

                case "ProximityList":
                    string[] tmp = new string[node.ChildNodes.Count];
                    type = TermType.Exact;
                    for (int i = 0; i < node.ChildNodes.Count; i++)
                    {
                        tmp[i] = ConvertQuery(node.ChildNodes[i], type);
                    }
                    result = "(" + string.Join(" NEAR ", tmp) + ")";
                    type = TermType.Inflectional;
                    break;

                case "Phrase":
                    result = '"' + node.Token.ValueString + '"';
                    break;

                case "ThesaurusExpression":
                    result = " FORMSOF (THESAURUS, " +
                        node.ChildNodes[1].Token.ValueString + ") ";
                    break;

                case "ExactExpression":
                    result = " \"" + node.ChildNodes[1].Token.ValueString + "\" ";
                    break;

                case "Term":
                    switch (type)
                    {
                        case TermType.Inflectional:
                            result = node.Token.ValueString;
                            if (result.EndsWith("*"))
                                result = "\"" + result + "\"";
                            else
                                result = " FORMSOF (INFLECTIONAL, " + result + ") ";
                            break;
                        case TermType.Exact:
                            result = node.Token.ValueString;

                            break;
                    }
                    break;

                // This should never happen, even if input string is garbage
                default:
                    throw new ApplicationException("Converter failed: unexpected term: " +
                        node.Term.Name + ". Please investigate.");

            }
            return result;
        }
        #endregion
    }
}
