﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using searchComponent.structure;
using Irony.Parsing;
using searchComponent.grammar;

namespace searchComponent.interpreter
{
    /// <summary>
    /// This class will form the sql query for the Numeric columns
    /// </summary>
    public class NumericColumnInterpreter : ColumnInterpreter
    {
        #region Constructor
        /// <summary>
        /// Constructs a NumericColumnInterpreter linked to a specific column and ready to 
        /// interpret a specific query.
        /// </summary>
        /// <param name="objColumn"></param>
        public NumericColumnInterpreter(Column objColumn)
        {
            this._column = objColumn;
        }
        #endregion

        #region Methods
        /// <summary>
        /// Gets the query needed for this column. THe node is present as a parameter
        /// because the method is recursive.
        /// </summary>
        /// <param name="node"></param>
        /// <returns></returns>
        public override string GetWhereQuery(string columnSearch)
        {
            NumericGrammar grammar = new NumericGrammar();
            LanguageData language = new LanguageData(grammar);
            Parser parser = new Parser(language);
            ParseTree parseTree = parser.Parse(columnSearch);
            if (parseTree.Status == ParseTreeStatus.Error)
            {
                string messages = "";
                ParserMessageList listMessages = parseTree.ParserMessages;
                foreach (ParserMessage objMsg in listMessages)
                {
                    messages += objMsg.Message + "\n";
                }
                throw new Exception(messages);
            }
            ParseTreeNode root = parseTree.Root;
            string query = ConvertQuery(root);
            return query;
        }

        /// <summary>
        /// This method should convert the node tree into a sql query for the column passed
        /// as a parameter in the constructor of this object.
        /// </summary>
        /// <param name="node"></param>
        /// <returns></returns>
        private string ConvertQuery(ParseTreeNode node)
        {
            string result = "";
            // Note that some NonTerminals don't actually get into the AST tree, 
            // because of some Irony's optimizations - punctuation stripping and 
            // node bubbling. For example, ParenthesizedExpression - parentheses 
            // symbols get stripped off as punctuation, and child expression node 
            // (parenthesized content) replaces the parent ParExpr node (the 
            // child is "bubbled up").
            switch (node.Term.Name)
            {
                case "Expression":
                    result += ConvertQuery(node.ChildNodes[0]);
                    break;

                case "OrExpression":
                    result += "( " + ConvertQuery(node.ChildNodes[0]) + " )";
                    result += " OR (" + ConvertQuery(node.ChildNodes[2]) + " )";
                    break;

                case "AndExpression":
                    result += "( " + ConvertQuery(node.ChildNodes[0]) + " )";
                    result += " AND (" + ConvertQuery(node.ChildNodes[2]) + " )";
                    break;

                case "AndTerm":
                    result += ConvertQuery(node.ChildNodes[0]);
                    break;

                case "Operand":
                    result += ConvertQuery(node.ChildNodes[0]);
                    break;

                case "UnaryTerm":
                    // only number
                    if (node.ChildNodes.Count == 1) {
                        result += this._column.FullName + "=" + node.ChildNodes[0].Token.Text;
                        break;
                    }
                    // operation + number
                    result += this._column.FullName + " " +
                        GetComparisonOperator(node.ChildNodes[0].ChildNodes[0].Token.Text) + " " +
                        node.ChildNodes[1].Token.Text;
                    break;
                case "BinaryTerm":
                    result += this._column.FullName + " between " + node.ChildNodes[0].Token.Text +
                        " and " + node.ChildNodes[2].Token.Text;
                    break;
                // This should never happen, even if input string is garbage
                default:
                    result += ". Please investigate.";
                    break;
            }
            return result;
        }

        /// <summary>
        /// Retiurns the comparison operator to be used. Actually, all is needed is to
        /// return the same operator than the grammar and in the case of ! we have 
        /// to return &lt;&gt;
        /// </summary>
        /// <param name="p"></param>
        /// <returns></returns>
        private string GetComparisonOperator(string p)
        {
            if (p.Equals("!"))
                return "<>";
            return p;
        }
        #endregion
    }
}
