﻿using System;
using System.Configuration;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using log4net;
using searchComponent.structure;
using System.Data;
using searchComponent.interpreter;
using System.Data.SqlClient;

namespace searchComponent.search
{
    /// <summary>
    /// This class should be the only class needed to obtain results from the database
    /// according to a search query. The method used for this should be search(string)
    /// </summary>
    public class Searcher
    {
        private static readonly ILog log = LogManager.GetLogger("Standard");

        #region Fields
        /// <summary>
        /// The search strcutture to use with this instance of the search component.
        /// </summary>
        private SearchStructure _structure;
        /// <summary>
        /// THe connection string used, by default, the value is SearchConnectionString
        /// </summary>
        private string _connectionString;
        #endregion

        #region Attributes
        /// <summary>
        /// THe connection string used, by default, the value is 
        /// ConfigurationManager.ConnectionStrings["SearchConnectionString"].ConnectionString
        /// </summary>
        public string ConnectionString
        {
            get { return _connectionString; }
            set { this._connectionString = value; }
        }
        #endregion

        #region Constructor
        /// <summary>
        /// Constructor for the class, the only thing needed is the name of the configuration
        /// to search. This is needed because the component can be used to search in different
        /// tables and with different purposes.
        /// The connection string set is the default
        /// </summary>
        public Searcher(string cfg) : 
            this(SearchStructure.GetStructure(cfg, SearchStructure.DEFAULT_CONN_STRING))
        {
            _connectionString = ConfigurationManager.ConnectionStrings[SearchStructure.DEFAULT_CONN_STRING].ConnectionString;
        }

        /// <summary>
        /// Constructor for the class, the only thing needed is the name of the configuration
        /// to search. This is needed because the component can be used to search in different
        /// tables and with different purposes.
        /// </summary>
        /// <param name="cfg"></param>
        /// <param name="connString"></param>
        public Searcher(string cfg, string connString) :
            this(SearchStructure.GetStructure(cfg, connString)) 
        {
            _connectionString = connString;
        }

        /// <summary>
        /// This is used when the client has a subclass of SearchStructure where it is defined
        /// already the structure, table, joins and columns to use. Normally after this you should
        /// set the connection string because it is set to the default.
        /// </summary>
        /// <param name="obj"></param>
        public Searcher(SearchStructure obj)
        {
            ConnectionStringSettings conSetting =
                ConfigurationManager.ConnectionStrings[SearchStructure.DEFAULT_CONN_STRING];
            if (conSetting == null)
            {
                log.Warn("Connection String NOT SET, set SearchConnectionString or define the connectionString attribute");
            }
            else
            {
                this._connectionString =
                    ConfigurationManager.ConnectionStrings[SearchStructure.DEFAULT_CONN_STRING].ConnectionString;
            }
            this._structure = obj;
        }
        #endregion

        #region Methods
        /// <summary>
        /// This is the main method used for the searching.
        /// </summary>
        /// <param name="query"></param>
        /// <returns></returns>
        public SearchResult search(string txt)
        {
            string query = txt.Replace("&", "\n&").Replace("|", "\n|");
            if (!query.Contains("\n") && !query.Contains("="))
                query = "all=" + query;

            string msg = "An empty string will return an empty set";
            SearchResult result = new SearchResult(msg, new DataSet("empty"), "");

            if (String.IsNullOrEmpty(query))
            {
                log.Warn(msg);
                return result;
            }

            try
            {
                SqlServerInterpreter objInterpreter = new SqlServerInterpreter(this._structure);
                string sqlQuery = objInterpreter.GetQuery(query);
                DataSet dtsResult = GetDataSet(sqlQuery);
                result = new SearchResult("NO ERROR", dtsResult, sqlQuery);
            }
            catch (Exception err)
            {
                log.Error("Could not parse and interpret", err);
                result = new SearchResult(err.Message, new DataSet("empty"), "NO QUERY");
            }
            return result;
        }

        /// <summary>
        /// Returns a suggestion of the terms expected for the grammar defined by the structure
        /// field.
        /// </summary>
        /// <param name="query"></param>
        /// <returns></returns>
        public List<SuggestLine> suggest(string query)
        {
            SqlServerInterpreter objInterpreter = new SqlServerInterpreter(this._structure);
            return objInterpreter.GetSuggestion(query);
        }

        /// <summary>
        /// This method will obtain the dataset that results from executing the query
        /// directly to the database.
        /// </summary>
        /// <param name="sqlQuery"></param>
        /// <returns></returns>
        private DataSet GetDataSet(string sqlQuery)
        {
            SqlConnection con = null;
            DataSet dts = null;
            try
            {
                con = new SqlConnection(this._connectionString);
                con.Open();

                dts = new DataSet(_structure.Tables[0].Name);
                SqlCommand cmd = new SqlCommand(sqlQuery, (SqlConnection)con);
                SqlDataAdapter adapter = new SqlDataAdapter(cmd);
                int nbRows = adapter.Fill(dts, _structure.Tables[0].Name);

                con.Close();
                con.Dispose();
            }
            catch (Exception err)
            {
                log.Error("There was an error accessing data", err);
                throw err;
            }
            finally
            {
                if (con != null)
                {
                    con.Close();
                    con.Dispose();
                }
            }
            return dts;
        }
        #endregion
    }
}
