﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace searchComponent.structure
{
    /// <summary>
    /// This class contains the defeinition of a column. It can define whether the coilumn
    /// will  be searchable, if it will be contained in the standard search, the title of the
    /// column in the resulting dataset when obtaining results.
    /// </summary>
    public class Column
    {
        #region Enums
        /// <summary>
        /// The enumeration that validates the type of columns to be searched. For each of these
        /// types the model knows how to interpret the grammar needed for the search.
        /// </summary>
        public enum ColumnType
        {
            Numeric, Date, String, Fts
        };
        #endregion

        #region Fields
        /// <summary>
        /// The name of the table where this column is defined
        /// </summary>
        private string _tableName;
        /// <summary>
        /// The name of the column as it appears in the sql database
        /// </summary>
        private string _name;
        /// <summary>
        /// The title of the column for the resulting dataset, when constructing this is the
        /// same than NAME. It has to be overriden manually.
        /// </summary>
        private string _title;
        /// <summary>
        /// The searchKey that will be used in the grammar.
        /// </summary>
        private string _searchKey;
        /// <summary>
        /// the type of the column. This mapping is done manually, but normally the rules are:
        /// NUMERIC: for any int, double,float, etc.
        /// STRING: for any varchar, char, nvarchar, text, etc.
        /// FTS: when the column has been defined as searchable with Full text search
        /// DATE: when datetime or date.
        /// </summary>
        private ColumnType _type;
        /// <summary>
        /// Whether the column appears in the list of results or not
        /// </summary>
        private bool _appearInResults;
        /// <summary>
        /// Whether the column is searchable or not, sometimes this could be used when we want to
        /// show the column but it does not participate of the searching process
        /// </summary>
        private bool _searchable;
        /// <summary>
        /// Whether the column is used when no column is defined in the search query.
        /// </summary>
        private bool _appearInStandardSearch;
        #endregion

        #region Attributes
        /// <summary>
        /// The name of the column as it appears in the sql database
        /// </summary>
        public string Name
        {
            get { return this._name; }
        }
        /// <summary>
        /// Returns the full name of a column. Thios means tblname.colname
        /// </summary>
        public string FullName
        {
            get { return this._tableName + "." + this._name; }
        }
        /// <summary>
        /// The name of the column as it appears in the sql database
        /// </summary>
        public string Title
        {
            get { return this._title; }
        }
        /// <summary>
        /// The name of the column as it appears in the sql database
        /// </summary>
        public string SearchKey
        {
            get { return this._searchKey; }
        }
        /// <summary>
        /// The name of the column as it appears in the sql database
        /// </summary>
        public ColumnType Type
        {
            get { return this._type; }
        }
        /// <summary>
        /// Whether the column appears in the list of results or not
        /// </summary>
        public bool AppearInResults
        {
            get { return this._appearInResults; }
            set { this._appearInResults = value; }
        }
        /// <summary>
        /// Whether the column is searchable or not, sometimes this could be used when we want to
        /// show the column but it does not participate of the searching process
        /// </summary>
        public bool Searchable
        {
            get { return this._searchable; }
            set { this._searchable = value; }
        }
        /// <summary>
        /// Whether the column is used when no column is defined in the search query.
        /// </summary>
        public bool AppearInStandardSearch
        {
            get { return this._appearInStandardSearch; }
            set { this._appearInStandardSearch = value; }
        }
        #endregion

        #region Constructors
        /// <summary>
        /// The constructor for a column. By default a column WILL be displayed in the 
        /// results, will NOT be part of the standard search and IS searchable.
        /// </summary>
        /// <param name="name"></param>
        /// <param name="title"></param>
        /// <param name="searchkey"></param>
        /// <param name="aType"></param>
        public Column(string name, string title, string searchkey, ColumnType aType, string tbl)
        {
            this._name = name;
            this._title = title;
            this._searchKey = searchkey;
            this._type = aType;
            this._tableName = tbl;

            this._appearInResults = true;
            this._appearInStandardSearch = false;
            this._searchable = true;
        }
        #endregion
    }
}
