﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using log4net;
using System.Threading;
using System.Configuration;

namespace searchComponent.structure
{
    /// <summary>
    /// This class contains a search strcuture which consists of tables and joins. 
    /// Each table contains a number of columns.
    /// It is important to observe the order in which the tables are being added to the
    /// list because it is inthe same order that the join statements will be used to 
    /// construct the query.
    /// It is supposeed that all tables come from the same database.
    /// </summary>
    public class SearchStructure
    {
        private static readonly ILog log = LogManager.GetLogger("Standard");

        #region Static
        public static readonly string DEFAULT_CONN_STRING = "SearchConnectionString";

        /// <summary>
        /// This hastable contains all the configurations loaded from the databsae for searching.
        /// This allows recovering from memory the configurations needed to perform searches. To 
        /// allow rewriting o fthe configurations, the method reload is provided that will look 
        /// again in the database for the newest parameters for a given configuration.
        /// </summary>
        private static Dictionary<string, SearchStructure> _structures = 
            new Dictionary<string, SearchStructure>();

        /// <summary>
        /// This is the lock used by the monitor when obtaining a searchstructure with
        /// the GetStrcutre methods.
        /// </summary>
        private static readonly object lockSearch = new object();

        /// <summary>
        /// Theis method searches for the searchstructure given by the confgi string passed in parameter.
        /// The method uses lazy instantiation and therefore will load the configuration only when
        /// needed. Also, it uses the static hashtable and will not load it every time needed from 
        /// the database, instead, it will look for it in  memory.
        /// The method uses a lock to eb thread safe.
        /// </summary>
        /// <param name="cfg"></param>
        /// <param name="connString">The name of the connection string variable to read from the configuration</param>
        /// <returns>Returns the searchstructure to use for that configuration. If no config is
        /// found it returns null. Log will contain the error.</returns>
        public static SearchStructure GetStructure(string cfg, string connString)
        {
            lock (lockSearch)
            {
                SearchStructure objSearch = null;
                if (_structures.ContainsKey(cfg))
                {
                    objSearch = _structures[cfg];
                    return objSearch;
                }
                try
                {
                    objSearch = new SearchStructure(cfg, connString);
                    _structures.Add(cfg, objSearch);
                }
                catch (Exception err)
                {
                    log.Error("Not able to read config: " + cfg, err);
                    objSearch = null;
                }
                return objSearch;
            }
        }
        #endregion

        #region Fields
        /// <summary>
        /// The list of tables. Observe that the order of tables correspond to the order in which the
        /// query will form the FROM clause in the SQL query.
        /// </summary>
        protected List<TableStructure> _tables = new List<TableStructure>();
        /// <summary>
        /// The name of the variable that is declared asa a connection string in the configuration
        /// file of the application using this library.
        /// </summary>
        protected string _connString;

        #endregion

        #region Attributes
        /// <summary>
        /// The list of tables. Observe that the order of tables correspond to the order in which the
        /// query will form the FROM clause in the SQL query.
        /// </summary>
        public List<TableStructure> Tables
        {
            get { return this._tables; }
        }
        #endregion

        #region Constructor

        /// <summary>
        /// THis is an empty constructor needed to be able to subclass this class
        /// </summary>
        protected SearchStructure()
        {
            ConnectionStringSettings conSetting =
                ConfigurationManager.ConnectionStrings[DEFAULT_CONN_STRING];
            if (conSetting == null) 
            {
                log.Warn("Default con string NOT SET. Please set SearchConnectionString or use other constructor");
            } 
            else 
            {
                this._connString = conSetting.ConnectionString;
            }
        }

        /// <summary>
        /// The constructor should look for the connection string SearchConnectionString to connect
        /// to the tables that describe the structure. But if it doesn't find this connection string
        /// then it should use the one hard coded: localhost / Prueba user:sa passwd:aaaaaa.
        /// Method is private because it can only be accessed by the GetStructure method.
        /// </summary>
        /// <param name="cfg">The name of the config record in teh database</param>
        /// <param name="connString">The complete connection string</param>
        protected SearchStructure(string cfg, string connString) 
        {
            this._connString = connString;

            data.DSTableTableAdapters.SearchConfigTableAdapter adapterConfig =
                new searchComponent.data.DSTableTableAdapters.SearchConfigTableAdapter();
            adapterConfig.Connection = new System.Data.SqlClient.SqlConnection(_connString);
            data.DSTable.SearchConfigDataTable dtConfig = adapterConfig.GetData(cfg);

            if (dtConfig.Rows.Count != 1)
            {
                string msg = "Configuration " + cfg + " is not present in database or is multiple times defined";
                log.Error(msg);
                throw new Exception(msg);
            }
            LoadSearchStructureFromDB(dtConfig[0].guid);
        }
        #endregion

        #region Methods
        /// <summary>
        /// This method generates all the necessary tables, joins and columns from a database
        /// that will configure the search grammar to be used in the search. AFterwards, the component
        /// should be able to perform Sql queries against the defined set of tables and columns
        /// </summary>
        /// <param name="guid">The guid of the searchconfig record</param>
        private void LoadSearchStructureFromDB(Guid guidConfig)
        {
            data.DSTableTableAdapters.SearchTableTableAdapter adapterTbl =
                new searchComponent.data.DSTableTableAdapters.SearchTableTableAdapter();
            adapterTbl.Connection = new System.Data.SqlClient.SqlConnection(_connString);
            data.DSTable.SearchTableDataTable dtTable = adapterTbl.GetData(guidConfig);

            foreach (data.DSTable.SearchTableRow rowTable in dtTable.Rows)
            {
                LoadTableFromDB(rowTable);
            }
        }

        /// <summary>
        /// This method loads the table read in the DB and loads all the columns 
        /// related to that table. It loads also the join used for that table
        /// </summary>
        /// <param name="rowTable"></param>
        private void LoadTableFromDB(searchComponent.data.DSTable.SearchTableRow rowTable)
        {
            TableStructure tbl = new TableStructure(rowTable.name);

            data.DSTableTableAdapters.SearchColumnTableAdapter adapter =
                new searchComponent.data.DSTableTableAdapters.SearchColumnTableAdapter();
            adapter.Connection = new System.Data.SqlClient.SqlConnection(_connString);
            data.DSTable.SearchColumnDataTable dtColumn = adapter.GetData(rowTable.guid);

            foreach (data.DSTable.SearchColumnRow row in dtColumn.Rows)
            {
                Column.ColumnType colType = GetColumnType(row.columnType);
                Column col = new Column(row.name, row.title, row.searchKey, colType, tbl.Name);
                col.Searchable = row.searchable != 0;
                col.AppearInStandardSearch = row.standardSearch != 0;
                col.AppearInResults = row.appearInResults != 0;

                tbl.addColumn(col);
            }

            data.DSTableTableAdapters.SearchJoinTableAdapter adapterJoin =
                new searchComponent.data.DSTableTableAdapters.SearchJoinTableAdapter();
            adapterJoin.Connection = new System.Data.SqlClient.SqlConnection(_connString);
            data.DSTable.SearchJoinDataTable dtJoin = adapterJoin.GetData(rowTable.guid);
            if (dtJoin.Rows.Count > 1)
            {
                log.Error("Fix database, table " + tbl.Name + " cannot have more than one join");
                throw new Exception("Fix database, table " + tbl.Name + " cannot have more than one join");
            }

            JoinStatement objJoin = null;
            if (dtJoin.Rows.Count == 1)
            {
                TableStructure tblBase = _tables[0];
                objJoin = new JoinStatement(
                    tblBase.Name + "." + dtJoin[0].baseColumn,
                    tbl.Name + "." + dtJoin[0].joinColumn,
                    GetJoinType(dtJoin[0].joinType));
            }
            
            this.addTable(tbl, objJoin);
        }

        /// <summary>
        /// Returns the join type depending on the string sent
        /// </summary>
        /// <param name="p"></param>
        /// <returns></returns>
        private JoinStatement.JoinType GetJoinType(string p)
        {
            switch (p)
            {
                case "INNER":
                    return JoinStatement.JoinType.Inner;
                case "LEFT":
                    return JoinStatement.JoinType.Left;
                case "OUTER":
                    return JoinStatement.JoinType.Outer;
                case "RIGHT":
                    return JoinStatement.JoinType.Right;
                default:
                    log.Error("No Join defined by the name: " + p);
                    throw new Exception("Error in Join definition");
            }
        }

        /// <summary>
        /// This method converts from the strings STR, NUM, FTS and DAT to the corresponding
        /// column type values
        /// </summary>
        /// <param name="p">Can be STR, NUM, FTS and DAT</param>
        /// <returns>The type, if it does not find a type, there is an error</returns>
        private Column.ColumnType GetColumnType(string p)
        {
            if (string.IsNullOrEmpty(p))
            {
                log.Error("parameter cannot be null or empty, fix database");
                throw new Exception("parameter cannot be null or empty, fix database");
            }
            if (p.Equals("STR")) return Column.ColumnType.String;
            if (p.Equals("NUM")) return Column.ColumnType.Numeric;
            if (p.Equals("FTS")) return Column.ColumnType.Fts;
            if (p.Equals("DAT")) return Column.ColumnType.Date;

            log.Error("parameter can only be STR, NUM, FTS or DAT");
            throw new Exception("parameter can only be STR, NUM, FTS or DAT");
        }

        /// <summary>
        /// Adds a table to the search structure. The table will be added to the list of tables.
        /// The method will build a join from the two other parameters. The column parameters 
        /// cannot be null except when there are no tables in the structure.
        /// </summary>
        /// <param name="tbl">The table to add</param>
        /// <param name="join">The table.column that will be joined, this should typically
        /// the name of the first table in the list of tables and the id column of that table.
        /// The table.column in the table being added.</param>
        public void addTable(TableStructure tbl, JoinStatement join )
        {
            if (tbl == null)
            {
                log.Error("Table parameter should not be null");
                return;
            }

            if (_tables.Count > 0 && join == null)
            {
                log.Error("Join parameter should not be null when tables are present");
                return;
            }

            if (_tables.Count > 0 && !_tables[0].CheckStringColumn(join.BaseTableColumn))
            {
                log.Error("Base column is not well formed: " + join.BaseTableColumn + 
                    " and should be: " + _tables[0].Name + ".<column name>");
                return;
            }

            if (_tables.Count > 0 && !tbl.CheckStringColumn(join.JoinTableColumn))
            {
                log.Error("Join column is not well formed: " + join.JoinTableColumn + 
                    " and should be: " + tbl.Name + ".<column name>");
                return;
            }

            foreach (TableStructure objTable in _tables)
            {
                if (objTable.Name.Equals(tbl.Name))
                {
                    log.Error("Table " + tbl.Name + " has already been added");
                    return;
                }
            }

            tbl.Join = join;
            _tables.Add(tbl);
        }

        /// <summary>
        /// REturns a column definition for searching. First splits the parameter with 
        /// the dot operator. Then searches for the table named after the first
        /// parameter (left side of the dot) and then it searches for the column in
        /// the table named after the second parameter (right side of the dot).
        /// </summary>
        /// <param name="fullName"></param>
        /// <returns></returns>
        public Column GetColumn(string fullName)
        {
            string[] tableColumn = fullName.Split('.');
            if (tableColumn.Length != 2)
            {
                log.Error("Name should be in the form <table>.<column>");
                return null;
            }

            TableStructure tblFound = null;
            foreach (TableStructure tbl in _tables)
            {
                if (tbl.Name.Equals(tableColumn[0]))
                {
                    tblFound = tbl;
                    break;
                }
            }
            if (tblFound == null)
            {
                log.Error("Table " + tableColumn[0] + " not found");
                return null;
            }

            Column colFound = null;
            foreach (Column col in tblFound.Columns)
            {
                if (col.Name.Equals(tableColumn[1]))
                {
                    colFound = col;
                    break;
                }
            }
            if (colFound == null)
            {
                log.Error("Column " + tableColumn[1] + " not found");
                return null;
            }

            return colFound;
        }

        /// <summary>
        /// REturns a column definition for searching. THe parameter gives the search
        /// key term used in the grammar.
        /// </summary>
        /// <param name="fullName"></param>
        /// <returns></returns>
        public Column GetColumnBySearchKey(string searchKey)
        {
            foreach (TableStructure tbl in _tables)
            {
                foreach (Column col in tbl.Columns)
                {
                    if (col.SearchKey.Equals(searchKey))
                    {
                        return col;
                    }
                }
            }

            log.Error("Column with key " + searchKey + " not found");
            return null;
        }

        #endregion
    }
}
